/***************************************************************************
                          log.h  -
                             -------------------
    begin                : mi� jul 16 2003
    copyright            : (C) 2003 by Damian Serrano Garcia
    email                : dserrano@alumnos.upm.es
 ***************************************************************************/
#ifndef _LOG_H_
#define _LOG_H_

#include <stdio.h>
#include "semaphore.h"
// log struct

typedef struct log_st {
  FILE* file;
  semaphore sema_log;
} log_t;

log_t *init_log_cc (log_t *, int);
int write_in_log_cc (log_t *, char *, int);
int close_log (log_t *);
log_t *init_log_qm (log_t *, int);
int write_in_log_qm (log_t *, char *);
log_t *init_log_flag (log_t *, char *);
int write_in_log_flag (log_t *, char *);
#endif

