#include "declarations.h"
bool isSmallerId(T_Tid tid1, T_Tid tid2) {
  int cl1 = tid1 % 100, cl2 = tid2 % 100;
  int id1 = tid1 / 100; int id2 = tid2 / 100;

  if (cl1 != cl2)
    return cl1 < cl2;
  return id1 < id2;
}

bool isEqualId(T_Tid tid1, T_Tid tid2) {
  return tid1 == tid2;
}
