#include "threads.h"

int createThread(void* th, void *arg, int *attr) {
  pthread_t thread1;
  return (pthread_create (&thread1, NULL, (void *) th, (void*) attr));
}

int sleepThread(int microseconds) {
  return (usleep (microseconds));
}

int yieldThread() {
  return(usleep(0));
}
