// ensemble
#include "hot_sys.h"
#include "hot_error.h"
#include "hot_msg.h"
#include "hot_ens.h"
#include "hot_thread.h"

#ifdef _THREADS

int createThread (void *th, void *arg,int *attr);
int sleepThread (int microseconds);
int yieldThread ();

#define _THREADS
#endif
