#ifndef _UTILITIES
#define _UTILITIES

#include <string.h>
#include "declarations.h"

void convertToInt (char *, int, int *, int);
void convertToChar (char *, int, int, int);
void typeQuery (char *, int *, int);
void tableAcceded (char *, char **);
void testTableAcceded ();
char * mayusculas (char *);
void itoa (int, char []);
void reverse (char s[]);
int  getToken (char *, char, int, char *);

#endif
