#ifndef _COMMITLISTH
#define _COMMMITLISTH

#include "transmng.h"

typedef struct T_CommitList  *T_CommitList_ptr;
 struct T_CommitList{
  //T_Tid  item;
  Commit_M data;
  T_CommitList_ptr next;
 };

T_CommitList_ptr CL_createlist();
T_CommitList_ptr CL_empty(T_CommitList_ptr);
T_CommitList_ptr CL_insertInOrder(T_CommitList_ptr, /*T_Tid,*/ Commit_M);
int CL_isIn(T_CommitList_ptr, T_Tid);
T_CommitList_ptr CL_remove(T_CommitList_ptr);
T_CommitList_ptr CL_removeByID(T_CommitList_ptr, T_Tid);
T_Tid CL_getFirstID(T_CommitList_ptr);
T_CommitList_ptr CL_getFirst(T_CommitList_ptr list);
//T_Transaction* CL_getTransaction(T_CommitList_ptr, T_Tid);


#endif
