// ensemble
#include "hot_sys.h"
#include "hot_error.h"
#include "hot_msg.h"
#include "hot_ens.h"
#include "hot_thread.h" 


#ifndef _SEMAPHORE

typedef hot_sema_t semaphore;

int createSemaphore (int initValue, semaphore *s);

int decSemaphore (semaphore *s);

int incSemaphore (semaphore *s);

#define _SEMAPHORE
#endif
