#ifndef _LISTH
#define _LISTH

#include "transmng.h"

typedef struct T_List  *T_List_ptr;
 struct T_List{
  T_Transaction*  item;
  T_List_ptr next;
 };

T_List_ptr L_createlist();
T_List_ptr L_empty(T_List_ptr);
T_List_ptr L_insertInOrder(T_List_ptr, T_Transaction*);
T_List_ptr L_remove(T_List_ptr);
T_List_ptr L_removeByID(T_List_ptr, T_Tid);
T_Transaction* L_getFirst(T_List_ptr);
T_Transaction* L_getTransaction(T_List_ptr, T_Tid);


#endif
